from mesa_designer.mesa import *

def test_annotation():
    a = Annotation("test", 5, 10, part_type="CDS")

    ap = AnnotatedPart("ABCDEFGHIJKLMNOPQRSTUVWXYZ", name="test_part")

    ap.add_annotation(a)

    ap.save_genbank_file("test.gb")


from __future__ import annotations
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from io import StringIO


class Annotation:
    def __init__(self, name: str, start: int | None=None, stop: int | None=None, part_type: str | None=None) -> None:
        self.name: str = name
        self.start: int = start if start else 0
        self.stop: int = stop if stop else len(name)
        self.type: str | None = part_type if part_type else "CDS"


    def shift_annotation(self, amount: int) -> Annotation:
        self.start += amount
        self.stop += amount

        return self


    def to_seq_feature(self) -> SeqFeature:
        return SeqFeature(location=FeatureLocation(self.start, self.stop), type=self.type, qualifiers={"name": self.name})


class AnnotatedPart:
    def __init__(self, sequence: str, name: str, part_id: str | None=None, description: str | None=None, seq_annotations: list[Annotation] | None=None) -> None:
        self.sequence: str = sequence if sequence else ""
        self.annotations: list[Annotation] = seq_annotations if seq_annotations else []
        self.name: str = name if name else "no_name"
        self.part_id: str = part_id if part_id else name
        self.description: str = description if description else name


    def get_sequence(self) -> str:
        return self.sequence


    def get_annotations(self) -> list[Annotation]:
        return self.annotations


    def add_annotation(self, annotation: Annotation) -> AnnotatedPart:
        if self.annotations is None:
            self.annotations = [annotation]
            return self

        self.annotations.append(annotation)

        return self


    def add_annotations(self, seq_annotations: list[Annotation]) -> AnnotatedPart:
        if self.annotations is None:
            self.annotations = seq_annotations
            return self

        self.annotations += seq_annotations

        return self

    def add(self, sequence: str, annotation: Annotation | None=None) -> AnnotatedPart:
        if not self.sequence:
            self.sequence = ""

        if not self.annotations:
            self.annotations = []

        self.sequence += sequence
        if annotation:
            self.annotations.append(annotation)

        return self


    def get_seq_record(self) -> SeqRecord:
        return SeqRecord(Seq(self.sequence), id=self.part_id, name=self.name, description=self.description, features=[
            annotation.to_seq_feature() for annotation in self.annotations], annotations={"molecule_type": "PROTEIN"})


    def to_genbank_string(self) -> str:
        f: StringIO = StringIO()
        SeqIO.write(self.get_seq_record(), f, "genbank")
        return f.getvalue()


    def save_genbank_file(self, file_path: str) -> None:
        content: str = self.to_genbank_string()
        with open(file_path, "w") as f:
            f.write(content)



from __future__ import annotations
from .part import AnnotatedPart, Annotation
from mesa_designer import TMD_DATA, AIP_DATA, SIGNAL_SEQS, NTEV_DATA, CTEV_DATA, TEVP_DATA, PRS_DATA

MESA_ORDER: list[str] = ["signal_peptide",
                         "tags",
                         "binder",
                         "tmd_linker",
                         "tmd",
                         "fret",
                         "protease",
                         "prs",
                         "cargo",
                         "aip"]


class MesaChain:
    def __init__(self) -> None:
        self.parts: dict[str, AnnotatedPart] = {}

    def get_parts(self) -> dict[str, AnnotatedPart]:
        return self.parts

    def add_binder(self, sequence: str, name: str | None = None, annotation: str | None = None) -> MesaChain:
        if not sequence:
            raise ValueError("Sequence cannot be None")

        self.parts["binder"] = AnnotatedPart(sequence=sequence,
                                             name=name if name else "Binder",
                                             seq_annotations=[Annotation("Binder" if not annotation else annotation)])

        return self

    def add_tmd_linker(self, sequence: str | None = None, name: str | None = None,
                       annotation: str | None = None) -> MesaChain:
        if not sequence:
            sequence = "GGGS" * 10

        self.parts["tmd_linker"] = AnnotatedPart(sequence=sequence,
                                                 name=name if name else "TMD Linker",
                                                 seq_annotations=[Annotation("Linker" if not annotation else annotation)])

        return self

    def add_protease_sequence(self, protease_name: str) -> MesaChain:
        proteases = NTEV_DATA
        proteases.update(CTEV_DATA)
        proteases.update(TEVP_DATA)

        if protease_name not in proteases.keys():
            raise ValueError(
                f"{protease_name} is not a valid Protease name. Please only use available NTEV, CTEV or TEVP names")

        self.parts["protease"] = AnnotatedPart(sequence=proteases[protease_name],
                                               name=protease_name,
                                               seq_annotations=[Annotation(name=protease_name)])

        return self

    def add_custom_protease_sequence(self, sequence: str, name: str | None = None,
                                     annotation: str | None = None) -> MesaChain:
        if not sequence:
            raise ValueError("Sequence cannot be None")

        self.parts["protease"] = AnnotatedPart(sequence=sequence,
                                               name=name if name else "Protease",
                                               seq_annotations=[Annotation("Protease" if not annotation else annotation)])

        return self

    def add_prs(self, prs_name: str | None = None) -> MesaChain:
        if not prs_name:
            prs_name = "PRS"

        if prs_name not in PRS_DATA.keys():
            raise ValueError(f"{prs_name} is not a valid PRS name. Please only use available PRS names")

        self.parts["prs"] = AnnotatedPart(sequence=PRS_DATA[prs_name][1],
                                          name=prs_name,
                                          seq_annotations=[Annotation(prs_name)])

        return self

    def add_custom_prs(self, sequence: str, name: str | None = None, annotation: str | None = None) -> MesaChain:
        if not sequence:
            raise ValueError("Protease Recognition Sequence cannot be None")

        self.parts["prs"] = AnnotatedPart(sequence=sequence,
                                          name=name if name else "PRS",
                                          seq_annotations=[Annotation(name="PRS" if not annotation else annotation)])

        return self

    def add_target(self, sequence: str, name: str | None = None, annotation: str | None = None) -> MesaChain:
        if not sequence:
            raise ValueError("Cargo Sequence cannot be None")

        self.parts["cargo"] = AnnotatedPart(sequence=sequence,
                                            name=name if name else "Cargo",
                                            seq_annotations=[Annotation(name="Cargo" if not annotation else annotation)])

        return self

    def add_tmd(self, tmd_name: str) -> MesaChain:
        tmd_name = tmd_name.upper()
        if not tmd_name in TMD_DATA.keys():
            raise ValueError(
                f"{tmd_name} is not a valid TMD name. Please only use available TMD names or use a custom TMD")

        self.parts["tmd"] = AnnotatedPart(sequence=TMD_DATA[tmd_name][1],
                                          name=f"{tmd_name}_TMD",
                                          seq_annotations=[Annotation(f"{tmd_name}_TMD")])

        return self

    def add_custom_tmd(self, sequence: str, name: str | None = None, annotation: str | None = None) -> MesaChain:
        if not sequence:
            raise ValueError("TMD Sequence cannot be None")

        self.parts["tmd"] = AnnotatedPart(sequence=sequence,
                                          name=name if name else "TMD",
                                          seq_annotations=[Annotation(f"{name}_TMD" if not annotation else annotation)])

        return self

    def add_signal_peptide(self, peptide_name: str | None = None) -> MesaChain:
        if not peptide_name:
            peptide_name = "CD4"

        peptide_name = peptide_name.upper()
        if peptide_name not in SIGNAL_SEQS.keys():
            raise ValueError(
                f"{peptide_name} is not a valid peptide name. Please only use available peptide names or use a custom peptide sequence")

        self.parts["signal_peptide"] = AnnotatedPart(sequence=SIGNAL_SEQS[peptide_name][1],
                                                     name=f"{peptide_name}_Signal_Peptide",
                                                     seq_annotations=[Annotation(f"{peptide_name}_Signal_Peptide")])

        return self

    def add_custom_signal_peptide(self, sequence: str, name: str | None = None,
                                  annotation: str | None = None) -> MesaChain:
        if not sequence:
            raise ValueError("Peptide sequence cannot be None")

        self.parts["signal_peptide"] = AnnotatedPart(sequence=sequence,
                                                     name=name if name else "Signal_Peptide",
                                                     seq_annotations=[Annotation(f"{annotation}_Signal_Peptide" if not annotation else annotation)])

        return self

    def add_aip(self, aip_name: str) -> MesaChain:
        aip_name = aip_name.upper()
        if not aip_name in AIP_DATA.keys():
            raise ValueError(
                f"{aip_name} is not a valid AIP name. Please only use available AIP names or use a custom AIP")

        self.parts["aip"] = AnnotatedPart(sequence=AIP_DATA[aip_name][1],
                                          name=f"{aip_name}_AIP",
                                          seq_annotations=[Annotation(f"{aip_name}_AIP")])

        return self

    def add_custom_aip(self, sequence: str, name: str, annotation: str | None = None) -> MesaChain:
        if not sequence:
            raise ValueError("AIP Sequence cannot be None")

        self.parts["aip"] = AnnotatedPart(sequence=sequence,
                                          name=name,
                                          seq_annotations=[Annotation(f"{name}_AIP" if not annotation else annotation)])

        return self

    def remove_component(self, component: str) -> MesaChain:
        if not component:
            raise ValueError("Component cannot be None")

        if component in self.parts.keys():
            self.parts.pop(component)

        else:
            raise ValueError("Component not in current MESA Chain")

        return self

    def to_annotated_part(self, name: str, part_id: str | None=None, description: str | None=None) -> AnnotatedPart:
        sequence: str = ""
        seq_annotations: list[Annotation] = []
        for component in MESA_ORDER:
            if component in self.parts.keys():
                seq_annotations.append(self.parts[component].get_annotations()[0].shift_annotation(len(sequence)))
                sequence += self.parts[component].get_sequence()

                # conditionally add short linker
                match component:
                    case "binder":
                        if "tmd_linker" not in self.parts.keys():
                            sequence += "GGGSGGGS"

                    case "tmd":
                        sequence += "GGGSGGGS"

                    case "protease":
                        if "prs" in self.parts.keys() or "cargo" in self.parts.keys() or "aip" in self.parts.keys():
                            sequence += "GGGSGGGS"

                    case "prs":
                        if "cargo" in self.parts.keys():
                            sequence += "GGGSGGGS"

                    case "cargo":
                        if "aip" in self.parts.keys():
                            sequence += "GGGSGGGS"

                        sequence += "*"

        if not sequence.startswith("M"):
            sequence = "M" + sequence
            for annotation in seq_annotations:
                annotation.shift_annotation(1)

        return AnnotatedPart(sequence=sequence,
                             name=name,
                             part_id=part_id if part_id else name,
                             description=description if description else name,
                             seq_annotations=seq_annotations)

    def to_genbank_string(self) -> str:
        return self.to_annotated_part(name="mesa_chain").to_genbank_string()

    def save_genbank_file(self, file_path: str) -> None:
        self.to_annotated_part(name="mesa_chain").save_genbank_file(file_path)


class MesaAssembly:
    def __init__(self, mesa_chains: dict[str, MesaChain] | None):
        self.mesa_chains: dict[str, MesaChain] = mesa_chains if mesa_chains else {}

    def set_chain(self, name: str, mesa_chain: MesaChain) -> MesaAssembly:
        if not (name and mesa_chain):
            raise ValueError("MESA Chain cannot be None")

        self.mesa_chains[name] = mesa_chain

        return self


if __name__ == "__main__":
    m = MesaChain()
    m.add_tmd("FGFR4")



import json
from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent
DATA_DIR = BASE_DIR / "data"

with open(DATA_DIR / "tmd/tmd_list.json", "r") as f:
    TMD_DATA: dict = dict(json.load(f))

with open(DATA_DIR / "aip/aip_list.json", "r") as f:
    AIP_DATA: dict = dict(json.load(f))

with open(DATA_DIR / "FRET/ICDs.json", "r") as f:
    FRET_ICDs: dict = dict(json.load(f))

with open(DATA_DIR / "intracellular/CTEV_list.json", "r") as f:
    CTEV_DATA: dict = dict(json.load(f))

with open(DATA_DIR / "intracellular/NTEV_list.json", "r") as f:
    NTEV_DATA: dict = dict(json.load(f))

with open(DATA_DIR / "intracellular/TEVp_list.json", "r") as f:
    TEVP_DATA: dict = dict(json.load(f))

with open(DATA_DIR / "prs/prs_list.json", "r") as f:
    PRS_DATA: dict = dict(json.load(f))

with open(DATA_DIR / "signal_seqs/signal_sequences.json", "r") as f:
    SIGNAL_SEQS: dict = dict(json.load(f))

with open(DATA_DIR / "tags/tag_sequences.json", "r") as f:
    TAG_SEQS: dict = dict(json.load(f))


from mesa_designer import TMD_DATA, AIP_DATA, FRET_ICDs, CTEV_DATA, NTEV_DATA, TEVP_DATA, PRS_DATA, SIGNAL_SEQS, TAG_SEQS

def test_data_imports():
    assert len(TMD_DATA) > 0
    assert len(AIP_DATA) > 0
    assert len(FRET_ICDs) > 0
    assert len(CTEV_DATA) > 0
    assert len(NTEV_DATA) > 0
    assert len(TEVP_DATA) > 0
    assert len(PRS_DATA) > 0
    assert len(SIGNAL_SEQS) > 0
    assert len(TAG_SEQS) > 0
